#include <SPI.h>
#include <Ethernet.h>
#include <Wire.h>
#include <Adafruit_NFCShield_I2C.h>

// NFC-definitions
#define IRQ   (2)
#define RESET (3)  // Not connected by default on the NFC Shield
Adafruit_NFCShield_I2C nfc(IRQ, RESET);

// Enter a MAC address and IP address for your controller below.
// The IP address will be dependent on your local network.
// gateway and subnet are optional:
byte mac[] = { 
  0x86, 0x18, 0xC8, 0x11, 0xFE, 0xDE };
IPAddress ip(192,168,0,111);

EthernetServer server(23);

String inputString = "";         // a string to hold incoming data
boolean stringComplete = false;  // whether the string is complete
boolean allIsOn = false;
uint8_t mifare_keya[6] = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
uint8_t mifare_secretBlock = 1;
uint8_t secret[16] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
char telnetSecret[] = "password";
uint8_t readData[16];

void setup() { 
  
  // Initialize pins:
  pinMode(22, OUTPUT);
  pinMode(30, OUTPUT);
  pinMode(32, OUTPUT);
  pinMode(36, OUTPUT);
  pinMode(38, OUTPUT);
  pinMode(42, OUTPUT);
  pinMode(44, OUTPUT);
  digitalWrite(22, LOW);
  digitalWrite(30, LOW);
  digitalWrite(32, LOW); 
  digitalWrite(36, LOW);
  digitalWrite(38, LOW); 
  digitalWrite(42, LOW);
  digitalWrite(44, LOW);
  
  // initialize serial:
  Serial.begin(115200);
  while(!Serial) {
    ; // Wait for serial
  }
  
  // reserve 200 bytes for the inputString:
  inputString.reserve(200);
    
  // initialize ethernet
  Ethernet.begin(mac, ip);
  ip = Ethernet.localIP();
  server.begin();
  
  // Initialize NFC:
  nfc.begin();
  Serial.println("NFC.begin");
  uint32_t versiondata = nfc.getFirmwareVersion();
  Serial.println("nfc.firmwareVersion");
  if (! versiondata) {
    while(!Serial) {
      ; // Wait for serial
    }
    Serial.println("Didn't find PN53x board");
    while (1); // halt
  }
  nfc.SAMConfig();
  Serial.println("nfc.SAMConfig");
  
}

uint32_t loopcounter = 0;
void loop() {
  loopcounter++;
  
  // Listen for NFC tag:
  if(loopcounter % 100 == 0) {
    listenForNFCtag();
  }

  
  // Telnet server:
  EthernetClient client = server.available();
  if(client) {
    char c;
    inputString = "";
    while((c = client.read()) != -1) {
      inputString += c;
      // if the incoming character is a newline, set a flag
      // so the main loop can do something about it:
      if (c == '\n' || inputString.length() == 199) {
        stringComplete = true;
        break;
      }
    }
  }
  
  
  // print the string when a newline arrives:
  if (stringComplete) {
    inputString.trim();
    if(authenticateTelnetInput(inputString)) {
      inputString = inputString.substring(16);
      inputString.toUpperCase();
      if(Serial) {
        Serial.println(inputString);
      }
      boolean isValid = getIsValid(inputString);
      String msg;
      if(isValid) {
        int outletNr = getOutletNr(inputString);
        boolean onState = getOnState(inputString);
        setOutletState(outletNr, onState);
        
        String stateString = (onState) ? "ON" : "OFF";
        String outletNrString = String(outletNr);
        msg = "Turned outlet " + outletNrString + " " + stateString;
      }
      else {
        msg = "Invalid command. Format has to be <OutletNr>-<State>, fx. 1-ON";
      }
      // Write message:
      if(Serial) {
        Serial.println(msg);
      }
      if(client) {
        msg += "\n";
        char msgBytes[msg.length()+1];
        msg.toCharArray(msgBytes,msg.length()+1);
        for(int i = 0; i < msg.length()+1; i++) {
          server.write(msgBytes[i]);
        }
      }
    }// if authenticateTelnetInput
    // clear the string:
    inputString = "";
    stringComplete = false;
  }
}

/*
  SerialEvent occurs whenever a new data comes in the
 hardware serial RX.  This routine is run between each
 time loop() runs, so using delay inside loop can delay
 response.  Multiple bytes of data may be available.
 */
void serialEvent() {
  while (Serial.available()) {
    // get the new byte:
    char inChar = (char)Serial.read(); 
    // add it to the inputString:
    inputString += inChar;
    // if the incoming character is a newline, set a flag
    // so the main loop can do something about it:
    if (inChar == '\n' || inputString.length() == 199) {
      stringComplete = true;
      break;
    } 
  }
}

void listenForNFCtag() {
  // NFC
  uint8_t success = 0x0;
  uint8_t uid[] = { 0, 0, 0, 0, 0, 0, 0 };  // Buffer to store the returned UID
  uint8_t uidLength;                        // Length of the UID (4 or 7 bytes depending on ISO14443A card type)

  // Wait for an ISO14443A type cards (Mifare, etc.).  When one is found
  // 'uid' will be populated with the UID, and uidLength will indicate
  // if the uid is 4 bytes (Mifare Classic) or 7 bytes (Mifare Ultralight)
  success = nfc.readPassiveTargetID(PN532_MIFARE_ISO14443A, uid, &uidLength,500);
  
  if (success) {
    if(Serial) {
       Serial.println("Tag read!");
    }
    success = nfc.mifareclassic_AuthenticateBlock (uid, uidLength, mifare_secretBlock, 0, mifare_keya);
    if (success) {
      if(Serial) {
        Serial.println("Tag authenticated!");
      }
      success = nfc.mifareclassic_ReadDataBlock(mifare_secretBlock, readData);
      if (success)
      {
        success = true;
        for(uint8_t i=0;i<16;i++) {
          if(secret[i] != readData[i]) {
            success = false;
          }
        }
        if(success) {
          if(Serial) {
            Serial.println("Secret confirmed!");
          }
          if(!allIsOn) {
            setOutletState(1, true);
            setOutletState(2, true);
            setOutletState(3, true);
            setOutletState(4, true);
            allIsOn = true;
            if(Serial) {
              Serial.println("Everything turned on!");
            }
          }
          else {
            setOutletState(1, false);
            setOutletState(2, false);
            setOutletState(3, false);
            setOutletState(4, false);
            allIsOn = false;
            if(Serial) {
              Serial.println("Everything turned off!");
            }
          }
          delay(3000);
        } // if secret confirmed
      } // if data read
    } // if block authenticated
  } // if tag read
}

boolean authenticateTelnetInput(String str) {
  boolean success = true;
  for(uint8_t i=0; i < 16; i++) {
    if(str.charAt(i) != telnetSecret[i]) {
      success = false;
    } 
  }
  return success;
}

boolean getIsValid(String str) {
  int dashPosition = str.indexOf("-");
  if(dashPosition < 1) {
    return false;
  }
 if(str.substring(0,dashPosition+1).toInt() == 0) {
   return false;
 }
 if(str.substring(dashPosition+1) != "ON" && str.substring(dashPosition+1) != "OFF") {
   return false;
 }
 return true; 
}

int getOutletNr(String str) {
  int dashPosition = str.indexOf("-");
  if(dashPosition < 1) {
    return 0;
  }
  return (int)str.substring(0,dashPosition).toInt();
}
boolean getOnState(String str) {
  int dashPosition = str.indexOf("-");
  if(dashPosition < 1) {
    return false;
  }
  if(str.substring(dashPosition+1) == "ON") {
    return true;
  }
  return false;
}

void setOutletState(int outlet, boolean onState) {
  // ###### ON: ######
  if(onState) {
    switch(outlet) {
      case 1:
        // DEV 1:
        digitalWrite(30, HIGH);
        delay(500);
        digitalWrite(30, LOW);
        break;
      case 2:
        // DEV 2:
        digitalWrite(36, HIGH);
        delay(500);
        digitalWrite(36, LOW);
        break;
      case 3:
        // DEV 3:
        digitalWrite(42, HIGH);
        delay(500);
        digitalWrite(42, LOW);
        break;
      case 4:
        // DEV 4:
        digitalWrite(22, HIGH);
        break;
    }
  }
  // ###### OFF: ######
  else {
    switch(outlet) {
      case 1:
        // DEV 1:
        digitalWrite(32, HIGH);
        delay(500);
        digitalWrite(32, LOW);
        break;
      case 2:
        // DEV 2:
        digitalWrite(38, HIGH);
        delay(500);
        digitalWrite(38, LOW);
        break;
      case 3:
        // DEV 3:
        digitalWrite(44, HIGH);
        delay(500);
        digitalWrite(44, LOW);
        break;
      case 4:
        // DEV 4:
        digitalWrite(22, LOW);
        break;
    }
  }
}
